#This is responsible for creating the .quill/project_documentation.md dir in the project repo

import subprocess
from pathlib import Path

def get_repo_root():
    """
    This is responsible for getting the main root of the project so that the .quill dir can get created there get_repo_root
    """

    try: 
        result = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            capture_output = True,
            text = True,
            check=True,
            
        )
        return Path(result.stdout.strip())
    except subprocess.CalledProcessError:
        return Path(".")

def create_quill_dir():
    """
    Responsible for craeting the .quill dir and the project-doc.md where all the documentation is written
    """

    root = get_repo_root()
    quill_dir = root /".quill"
    quill_dir.mkdir(exist_ok = True)

    doc_file = quill_dir / "project_documentation.md"
    if not doc_file.exists():
        doc_file.write_text("# Poject Documentation\n\n")



if __name__ == "__main__":
    create_quill_dir()


#Just a simple file to help me understand parsers

# Learning how to extract functions
def extract_functions(node, source_code):
    functions = [] # Collect the function names

    if node.type == "function_definition":
        name_node = node.child_by_field_name("name")

        if name_node:
            functionName = source_code[name_node.start_byte : name_node.end_byte].decode("utf-8")
            line_num = node.start_point[0] + 1

            functions.append({
                "name": functionName,
                "line": line_num
            })

    for child in node.children:
        functions.extend(extract_functions(child, source_code))
    
    return functions


def extract_classes(node, source_code):
    classes = [] # Collect the function names

    if node.type == "class_definition":
        name_node = node.child_by_field_name("name")

        #this should now allow me to get both the name of teh node liek class and where it lies in the file
        if name_node:
            className = source_code[name_node.start_byte : name_node.end_byte].decode("utf-8")
            line_num = node.start_point[0] + 1
            classes.append({
                "name" : className,
                "line" : line_num
            })

    for child in node.children:
        classes.extend(extract_classes(child, source_code))
    
    return classes




#Responsible for writing down the documentation into the file
#This page is also responsible for collecting all modified files using git then determine the correct parser
#Maybe make a seperate file like or folder like Project_Documenation/Docs.md or .txt
# this is the final page where it all comes together like a master-piece 

from datetime import datetime
import os 
import sys
import subprocess
from tree_sitter import Parser, Language
import tree_sitter_cpp as tscpp
import tree_sitter_javascript as tsjs
import tree_sitter_python as tspy
from rich.console import Console
import json
import re  # regex

# Calling my parsers
from . import python_parser, cpp_parser, js_parser


console = Console()



def get_resources_Path(relative_path):
    """
    This should get the respurces needed for quill
    """
    if hasattr(sys, '_MEIPASS'):
        return os.path.join(sys._MEIPASS, relative_path)
    
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), relative_path)


def get_current_time():
    """
    This should allow for quill to get the local time of every user
    """
    current_datetime_local = datetime.now()
    date_string = current_datetime_local.strftime("%Y-%m-%d %H:%M:%S")

    return date_string


def get_modified_files():
    """
    This function gets all the modified files that the dev worked on in the project so far 
    """
    try:
        #checking if HEAD exists to prevent crash on new repos
        subprocess.run(["git", "rev-parse", "HEAD"], check = True, capture_output=True)

        commit_count = subprocess.check_output(["git", "rev-list", "--count", "HEAD"]).decode("utf-8").strip()
        commit_count = int(commit_count)
        if commit_count <= 1:
            #For when i dont have a commit in the tree yet, quill shouldnt crash like before
            command = ["git", "ls-tree", "-r", "HEAD", "--name-only"]
        else:
            command = ["git", "diff", "--name-only", "HEAD~1", "HEAD"]

        modified_files = subprocess.check_output(command).decode('utf-8').splitlines()
        return modified_files
    except:
        return []  # For when theres no commits
    
def pipeline(files):
    '''
    This is a mapping dictionary to map the correct file to parsing logic and it handles the 
    extraction of functions, classes and structs i got from the code 
    '''
    Language_Map = {
        '.py' : Language(tspy.language()),
        '.cpp' : Language(tscpp.language()),
        '.hpp' : Language(tscpp.language()),
        '.js' : Language(tsjs.language())
    }

    #storing all info
    all_extracted_info = []

    for file_path in files:
            ext = os.path.splitext(file_path)[1]
            if ext in Language_Map:
                lang = Language_Map[ext]
                parser = Parser(lang)

                #creating a dictionary for this file
                file_info = {"file": file_path, "functions": [], "classes": [], "structs": []}

                if os.path.exists(file_path):
                     with open(file_path, "rb") as f:
                          source_code = f.read()     # reads the source code as bytes since tree sitter need bytes
                     tree = parser.parse(source_code)
                     root = tree.root_node
                    
                     if ext == ".py":
                          file_info["functions"] = python_parser.extract_functions(root, source_code)
                          file_info["classes"] = python_parser.extract_classes(root, source_code)
                          
                     
                     elif ext == ".cpp" or ext == ".hpp" :
                          file_info["structs"] = cpp_parser.extract_structs(root, source_code)
                          file_info["functions"] = cpp_parser.extract_functions(root, source_code)
                          file_info["classes"] = cpp_parser.extract_classes(root, source_code)
                          
                     
                     elif ext == ".js":
                          file_info["classes"] = js_parser.extract_classes(root, source_code)
                          file_info["functions"] = js_parser.extract_functions(root, source_code)
                          
                     all_extracted_info.append(file_info)
    return all_extracted_info


def code_summary(data_list):
     """
     After the commit to get this should appear to the dev showing them all their modified files
     and all the added functions, classes and in cpp structs
     """
     indicolite = "#267D97" # Colour code of indicolit i could find
     console.print("Quill Code Summary : " , style = f"bold italic underline {indicolite}", justify = "center")

     for item in data_list:
          console.print(f"\nFile : {item['file']}", style = "green")

          if item['classes']:
        # Added (L{c['line']}) to show the line number in the terminal
            class_names = [f"{c['name']} (L{c['line']})" for c in item['classes'] ]
            console.print(f"  [bold green]Classes:[/bold green]   {', '.join(class_names)}")

          if item['functions']:
            
            function_names = [f"{f['name']} (L{f['line']})" for f in item['functions']]
            console.print(f"  [bold white]Functions: [/bold white]{', '.join(function_names)}")

          if item.get('structs'):
            struct_names = [f"{s['name']} (L{s['line']})" for s in item['structs']]
            console.print(f"  [bold magenta]Structs: [/bold magenta]   {', '.join(struct_names)}")

def reference_system(dev_notes, extracted_data):
    """
    This is responsible for allowing the user to use @ to link where this block of lives in teh system i
    in the project-doc.md file , creates a link to the file and block of code
    """
    refs = re.findall(r"@(\w+)", dev_notes)

    for ref in refs:
        # Find the matching data in your extracted list
        for file_entry in extracted_data:
            # Check functions, classes, and structs
            all_items = file_entry['functions'] + file_entry['classes'] + file_entry.get('structs', [])
            
            match = next((item for item in all_items if item['name'] == ref), None)
            
            if match:
                file_path = file_entry['file']
                line = match['line']
                # Replace @add with a Markdown link to that file and line
                link = f"[**{ref}**](../{file_path}#L{line})"
                dev_notes = dev_notes.replace(f"@{ref}", link)
                
    return dev_notes


def get_line():
    """
    This should help with the git hook . actually stop when the commit is ran
    """
    try:
        if os.name == 'nt':   #nt = Windows
            with open('CON', 'r') as f:
                return f.readline()
        else:
            with open('/dev/tty', 'r') as f:
                return f.readline()
    except Exception:
        return sys.stdin.readline()

def write_to_Docs(data_list):
    """
    This is responsible for writing all the information in to the project doc and index.json
    in the .quill file in the dir
    """
    if not data_list:
        return "No files we're modified."
    
    #What quill found
    code_summary(data_list)

    #notes from dev
    print("=> Add a quick note?")
    print("--Use @ to link important functions, classes or structs.")
    print("-- Press Ctrl + D or type DONE to save and exit.")
    dev_notes = []
    while True:
        try:
            line = get_line()
            if line.strip().upper() == "DONE":
                break
            dev_notes.append(line)
        except EOFError: # this should trigger Ctrl+D for a user to save and exit
            break
    
    #this should combine the strings of the dev notes nicely
    raw_notes = "\n".join(dev_notes)

    #now apply my reference code
    final_notes = reference_system(raw_notes, data_list)

    timestamp = get_current_time()
    os.makedirs(".quill", exist_ok=True)
    
    #md
    with open(".quill/project_documentation.md", "a", encoding="utf-8") as md_file:
        md_file.write(f"\n---\n## {timestamp}\n")
        md_file.write(f"Dev Notes:\n {final_notes}\n\n")
        for item in data_list:
            md_file.write(f"### `{item['file']}`\n")
            if item['classes']:
                class_names = [c['name'] for c in item['classes']]
                md_file.write(f"- Classes: {', '.join(class_names)}\n")
            if item['functions']:
                  # Extract names from the dictionaries
                func_names = [f['name'] for f in item['functions']]
                md_file.write(f"- Functions: {', '.join(func_names)}\n")
            md_file.write("\n")

    #json
    #No idea why im doing a json, but professional stuff uses .json to track
    json_path = ".quill/index.json"
    history = []
    if os.path.exists(json_path):
        with open(json_path, "r") as jf:
            try: history = json.load(jf)
            except: history = []

    history.append({
        "timestamp": timestamp,
        "note": final_notes,
        "files_modified": data_list
    })

    with open(json_path, "w") as jf:
        json.dump(history, jf, indent=4)


# This file is responsible for allowing Quill to run within your repo

import typer

from .git_check import check_git_initialized, check_git_installed
from .create_quill_dir import create_quill_dir
from .install_git_hook import install_hook
from .preflight import quill_preflight_check
from .write import get_modified_files, pipeline, write_to_Docs

app = typer.Typer(help = "Ace Quill - This is a tool that works with your git flow to keep documentation consistent")

@app.command() 
def init():
    # This should allow for users to run either quill or ace run to initialize Ace Doc in repo
    """
    Type in quill init to initilaze Ace Doc within your git repo. 
    """
    typer.echo("Initializing Quill ....")
    # TODO:
    check_git_installed()
    check_git_initialized()
    create_quill_dir()
    install_hook()
    typer.echo("Quill initialization completed.")

@app.command()
def run():
    """
    - Start running Quill in your repo.
    - If you're having trouble with quill run try quill init first to initialize quill into your git workflow.
    """
    #Building execution structure
    if quill_preflight_check():
        typer.echo("Quill is checking last commit....")
        #get files modified
        files = get_modified_files()
        # Extract and write
        extracted_data = pipeline(files)
        write_to_Docs(extracted_data)



def main():
    app()


if __name__ == "__main__":
    main()


# Checking to see if a device has git installed
import subprocess
import shutil
import platform

def check_git_installed():
    git_path = shutil.which("git")   # Checking for a git executable file in path
    if git_path:

        try:
            subprocess.check_output([
                "git", "--version"
            ],
            text = True,
            stderr = subprocess.DEVNULL
            )
            print(f" Git has been detected on your {platform.system()} device")
            return True
        except (subprocess.CalledProcessError, OSError):
            raise RuntimeError("Git was found but failed to execute")
            
    else:
        print(f"Git not found or installed on your {platform.system()} device. PLease install Git to use Quill ")

if __name__ == "__main__":
    check_git_installed()



# checking if git has been initialized in the project for quill to run

def check_git_initialized() -> bool:

    try: 
        subprocess.check_output(["git", "rev-parse", "--is-inside"],
                    
                    stderr = subprocess.DEVNULL           
        )
        return True
    except subprocess.CalledProcessError:
        print("Git has not been initilized , please exceute git init to get started with Quill")


#Responsible for extracting classes and functions/methods in js code


def extract_classes(node, source_code):
    classes = [] # Collect the function names

    if node.type == "class_declaration":
        name_node = node.child_by_field_name("name")

        if name_node:
            className = source_code[name_node.start_byte : name_node.end_byte].decode("utf-8")
            line_num = node.start_point[0] + 1
            classes.append({
                "name": className,
                "line" : line_num
            })

    for child in node.children:
        classes.extend(extract_classes(child, source_code))
    
    return classes

def extract_functions(node, source_code):
    functions = [] # Collect the function names

    if node.type == "method_definition":
        name_node = node.child_by_field_name("name")

        if name_node:
            functionName = source_code[name_node.start_byte : name_node.end_byte].decode("utf-8")
            line_num = node.start_point[0] + 1
            functions.append({
                "name": functionName,
                "line" : line_num
            })

    for child in node.children:
        functions.extend(extract_functions(child, source_code))
    
    return functions




#This is to enforce the rules of quill with git 
# To stick to its workflow for the best experience

from .git_check import check_git_initialized, check_git_installed
import typer


def quill_preflight_check():
    if not check_git_installed():
        typer.echo("Quill needs git to be installed")
        raise typer.Exit(code=1)

    if not check_git_initialized():
        typer.echo("Git is has not been initialized for this project")
        typer.echo("Run in project root : git init then quill init")
        raise typer.Exit(code=1)

    
    return True


# This is responsible for installing the git hooks onto a users device

import subprocess
from pathlib import Path
import shutil
import platform
from importlib.resources import files

def get_hook_path():
    """
    Get where the .git/hooks path is in the repo
    """
    result = subprocess.run(["git", "rev-parse", "--git-path", "hooks"],
                            capture_output = True,
                            text = True,
                            check = True,
                            
                            )
    return Path(result.stdout.strip())

def get_hook_script():
    """
    This should be responsible for getting the correct hook based on the OS .sh(Unix based) and .bat(Windows)
    """

    if platform.system() == "Windows":
        return "post-commit.bat"
    return "post-commit.sh"

def install_hook():
    """
    This function is responsible for installing the post-commit git hook onto the device
    """
    hooks_path = get_hook_path()
    script_name = get_hook_script()

    source = files("quill").joinpath("hooks", script_name)
    if platform.system() == "Windows":
        destination = hooks_path / "post-commit.bat"
    else:
        destination = hooks_path / "post-commit"

    if destination.exists():
        # Maybe offer to overwrite if you update Quill later!
        print("Post-Commit hook already exists.")
        return
    
    shutil.copy(source, destination)

    if platform.system() != "Windows":
        # Ensure it is executable
        destination.chmod(0o775)
        
    print(f"Quill git hook installed successfully as {destination.name}")